#include "common.h"
#include <9pclient.h>
#include "appendfile.h"

void
Fdassign(Fd* F, Ftype type, void* fd)
{
	F->type = type;
	switch(type){
	case Fcfd:
		F->cfd = (CFid*)fd;
		break;
	case Ffd:
		F->fd = *(int*)fd;
		break;
	}
}

int
Fdwrite(Fd* F, char* buf, int n){
	switch(F->type){
	case Fcfd:
		return fswrite(F->cfd, buf, n);
	case Ffd:
		return write(F->fd, buf, n);
	}
	return 0;
}

int
Fdread(Fd* F, char* buf, int n){
	switch(F->type){
	case Fcfd:
		return fsread(F->cfd, buf, n);
	case Ffd:
		return read(F->fd, buf, n);
	}
	return 0;
}

int
Fdseek(Fd* F, vlong offset, int whence){
	switch(F->type){
	case Fcfd:
		return fsseek(F->cfd, offset, whence);
	case Ffd:
		return seek(F->fd, offset, whence);
	}
	return 0;
}

static Inbuf*
allocinbuf(Fd* in, Fd* out)
{
	Inbuf *b;

	b = mallocz(sizeof(Inbuf), 1);
	if(b == nil)
		sysfatal("reading mailbox: %r");
	b->rp = b->wp = b->buf;
	b->in = in;
	b->out = out;
	return b;
}

static int
fill(Inbuf *b, int addspace)
{
	int i, n;

	if(b->eof && b->wp - b->rp == 0)
		return 0;

	n = b->rp - b->buf;
	if(n > 0){
		i = Fdwrite(b->out, b->buf, n);
		if(i != n)
			return -1;
		b->last = b->buf[n-1];
		b->bytes += n;
	}
	if(addspace){
		if(Fdwrite(b->out, " ", 1) != 1)
			return -1;
		b->last = ' ';
		b->bytes++;
	}

	n = b->wp - b->rp;
	memmove(b->buf, b->rp, n);
	b->rp = b->buf;
	b->wp = b->rp + n;

	i = Fdread(b->in, b->buf+n, sizeof(b->buf)-n);
	if(i < 0)
		return -1;
	b->wp += i;

	return b->wp - b->rp;
}

/* code to escape ' '*From' ' at the beginning of a line */
int
appendtombox(Inbuf* b)
{
	int addspace;
	int n;
	char *p;
	int sol;

	Fdseek(b->out, 0, 2);

	addspace = 0;
	sol = 1;

	for(;;){
		if(b->wp - b->rp < 5){
			n = fill(b, addspace);
			addspace = 0;
			if(n < 0)
				goto error;
			if(n == 0)
				break;
			if(n < 5){
				b->rp = b->wp;
				continue;
			}
		}

		/* state machine looking for ' '*From' ' */
		if(!sol){
			p = memchr(b->rp, '\n', b->wp - b->rp);
			if(p == nil)
				b->rp = b->wp;
			else{
				b->rp = p+1;
				sol = 1;
			}
			continue;
		} else {
			if(*b->rp == ' ' || strncmp(b->rp, "From ", 5) != 0){
				b->rp++;
				continue;
			}
			addspace = 1;
			sol = 0;
		}
	}

	/* mailbox entries always terminate with two newlines */
	n = b->last == '\n' ? 1 : 2;
	if(Fdwrite(b->out, "\n\n", n) != n)
		goto error;
	n += b->bytes;
	return n;
error:
	return -1;
}

int
append(Inbuf* b)
{
	int n;

	Fdseek(b->out, 0, 2);

	for(;;){
		n = fill(b, 0);
		if(n < 0)
			goto error;
		if(n == 0)
			break;
		b->rp = b->wp;
	}
	n = b->bytes;
	return n;
error:
	return -1;
}

int
appendFd(Fd* in, Fd* out)
{
	int r;
	Inbuf* b;

	b = allocinbuf(in, out);
	r = append(b);
	free(b);
	return r;
}

int
appendFdtombox(Fd* in, Fd* out)
{
	int r;
	Inbuf* b;

	b = allocinbuf(in, out);
	r = appendtombox(b);
	free(b);
	return r;
}

int
appendfiletombox(int fdin, int fdout)
{
	Fd in;
	Fd out;

	Fdassign(&in, Ffd, &fdin);
	Fdassign(&out, Ffd, &fdout);
	return appendFdtombox(&in, &out);
}
