#include "common.h"
#include <ndb.h>
#include "smtp.h"	/* to publish dial_string_parse */
#include <ip.h>

enum {
	Maxstring = 256,
};

static int	callmx(DS*, char*);
static void 	expand_meta(DS *ds);

extern int 	debug;

int
mxdial(char *addr, char *ddomain, char *gdomain)
{
	int fd;
	DS ds;
	char err[Errlen];

	addr = netmkaddr(addr, 0, "smtp");
	dial_string_parse(addr, &ds);

	/* try connecting to destination or any of it's mail routers */
	fd = callmx(&ds, ddomain);

	/* try our mail gateway */
	rerrstr(err, sizeof(err));
	if(fd < 0 && gdomain && strstr(err, "can't translate") != 0) {
		fprint(2,"dialing %s\n",gdomain);
		fd = dial(netmkaddr(gdomain, 0, "smtp"), 0, 0, 0);
	}

	strcpy(ddomain, ds.host);
	return fd;
}

/* 
  * we are trying to defend against people exportint addresses like
  * 127.0.0.1 in mx records. of course the silly way we're doing this
  * will not help us if somebody is trying to dos us with a mx record
  * like ::1. well, actually, ::1 won't bother us now, but addresses like
  * 192.168.0.0/16 and 10.0.0.0/8 will get us. - ewq.
  *
  */
static char * badip[] = {
	"127.0.0.1",		// sneaky.
	"::1",			// really sneaky
	0
};

static int
validmx(Ndbtuple* T)
{
	Ndbtuple* M;
	Ndbtuple* A;
	Ndbtuple* I;
	char** a;

	for (M=T; M = ndbfindattr(M, 0, "mx"); M = M->entry){
		A = dnsquery(0, M->val, "ip");
		if (!A)
			continue;
		for(I = A; I = ndbfindattr(I, 0, "ip"); I = I->entry) {
			if (debug)
				fprint(2, "check: %s %s\n", M->val, I->val);
			for(a=badip; *a; a++)  {
				if (!strcmp(I->val, *a)) {
					ndbfree(A);
					return 0;
				}
			}
		}
		ndbfree(A);
	}
	return 1;
}

/*
 *  take an address and return all the mx entries for it,
 *  most preferred first
 */
static int
callmx(DS* ds, char *ddomain)
{
	int fd;
	char addr[Maxstring];
	Ndbtuple* T;
	Ndbtuple* M;

	/* get a list of mx entries */
	T = dnsquery(0, ds->host, "mx");
	if(!T){
		if(debug)
			fprint(2, "mxlookup returns nothing\n");
		return dial(ddomain, 0, 0, 0);
	}

	/* refuse to honor loopback addresses given by dns */
	if (!validmx(T)) {
		ndbfree(T);
		if(debug)
			fprint(2, "illegal: domain %s lists nonpublic address\n", ds->host);
		werrstr("illegal: domain %s lists nonpublic address", ds->host);
		return -1;
	}

	/* dial each one in turn */
	for (M=T, fd=-1; fd < 0 && (M = ndbfindattr(M, 0, "mx")); M=M->entry){
		snprint(addr, sizeof(addr), "%s!%s!%s", ds->proto, M->val, ds->service);
		if(debug)
			fprint(2, "mxdial trying %s\n", addr);
		fd = dial(addr, 0, 0, 0);
	}
	ndbfree(T);
	return fd;
}

/* break up an address to its component parts */
void
dial_string_parse(char *str, DS *ds)
{
	char *p, *p2;

	strncpy(ds->buf, str, sizeof(ds->buf));
	ds->buf[sizeof(ds->buf)-1] = 0;

	p = strchr(ds->buf, '!');
	if(p == 0) {
		ds->proto = "net";
		ds->host = ds->buf;
	} else {
		if(*ds->buf != '/'){
			ds->proto = ds->buf;
		} else {
			for(p2 = p; *p2 != '/'; p2--)
				;
			*p2++ = 0;
			ds->proto = p2;
		}
		*p = 0;
		ds->host = p + 1;
	}
	ds->service = strchr(ds->host, '!');
	if(ds->service)
		*ds->service++ = 0;
	if(*ds->host == '$')
		expand_meta(ds);
}

#if 0 /* jpc */
static void
expand_meta(DS *ds)
{
	char buf[128], cs[128], *net, *p;
	int fd, n;

	net = ds->netdir;
	if(!net)
		net = "/net";

	if(debug)
		fprint(2, "expanding %s!%s\n", net, ds->host);
	snprint(cs, sizeof(cs), "%s/cs", net);
	if((fd = open(cs, ORDWR)) == -1){
		if(debug)
			fprint(2, "open %s: %r\n", cs);
		syslog(0, "smtp", "cannot open %s: %r", cs);
		return;
	}

	snprint(buf, sizeof(buf), "!ipinfo %s", ds->host+1);	// +1 to skip $
	if(write(fd, buf, strlen(buf)) <= 0){
		if(debug)
			fprint(2, "write %s: %r\n", cs);
		syslog(0, "smtp", "%s to %s - write failed: %r", buf, cs);
		close(fd);
		return;
	}

	seek(fd, 0, 0);
	if((n = read(fd, ds->expand, sizeof(ds->expand)-1)) < 0){
		if(debug)
			fprint(2, "read %s: %r\n", cs);
		syslog(0, "smtp", "%s - read failed: %r", cs);
		close(fd);
		return;
	}
	close(fd);

	ds->expand[n] = 0;
	if((p = strchr(ds->expand, '=')) == nil){
		if(debug)
			fprint(2, "response %s: %s\n", cs, ds->expand);
		syslog(0, "smtp", "%q from %s - bad response: %r", ds->expand, cs);
		return;
	}
	ds->host = p+1;

	/* take only first one returned (quasi-bug) */
	if((p = strchr(ds->host, ' ')) != nil)
		*p = 0;
}
#endif /* jpc */

static void
expand_meta(DS *ds)
{
	Ndb *db;
	Ndbs s;
	char *sys, *smtpserver;

	sys = sysname();
	db = ndbopen(0);
	fprint(2,"%s",ds->host);
	smtpserver = ndbgetvalue(db, &s, "sys", sys, "smtp", nil);
	snprint(ds->host,128,"%s",smtpserver);
	fprint(2," exanded to %s\n",ds->host);
	ndbclose(db);
}
